# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import reduce
from pathlib import Path
from typing import List, Optional, Tuple

import pandas as pd
import torch
import torch.nn.functional as F

from nncf.common.graph.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.logging import nncf_logger
from nncf.experimental.common.pruning.nodes_grouping import get_pruning_groups
from nncf.experimental.common.pruning.nodes_grouping import select_largest_groups
from nncf.experimental.common.pruning.propagation_data import ProducerInfo
from nncf.experimental.torch.pruning.operations import PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES
from nncf.experimental.torch.sparsity.movement.layers import MovementSparsifier
from nncf.experimental.torch.sparsity.movement.layers import SparseStructure
from nncf.torch.layers import NNCFLinear
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import SparseModuleInfo

SUPPORTED_NNCF_MODULES = [NNCFLinear]
EXPECTED_NODE_LAYER_ATTRS = [LinearLayerAttributes]


class StructuredMaskContextStatistics:
    """
    Describes details of the resolved structured mask in a supported layer.
    """

    def __init__(
        self,
        weight_shape: Tuple[int, int],
        pruned_weight_shape: Tuple[int, int],
        bias_shape: Tuple[int],
        pruned_bias_shape: Tuple[int],
        head_or_channel_id_to_keep: List[int],
        module_node_name: NNCFNodeName,
    ):
        """
        Initializes the statistics for the target linear module of a structured mask context.

        :param weight_shape: Shape of the original weight in a linear layer.
        :param pruned_weight_shape: Shape of the weight after structured mask resolution,
            discarding the pruned regions.
        :param bias_shape: Shape of the original bias in a linear layer.
        :param pruned_bias_shape: Shape of the bias after structured mask resolution,
            discarding the pruned regions.
        :param head_or_channel_id_to_keep: If the layer belongs to multi-head self-attention block,
            specifies the preserved head ids after structured masking. If the layer belongs to feed-
            forward network, specifies the preserved channels after structured masking.
        :param module_node_name: Node name of the target module.
        """
        self.weight_shape = weight_shape
        self.pruned_weight_shape = pruned_weight_shape
        self.bias_shape = bias_shape
        self.pruned_bias_shape = pruned_bias_shape
        self.head_or_channel_id_to_keep = head_or_channel_id_to_keep
        self.module_node_name = module_node_name


class StructuredMaskContext:
    """
    Context to interact with the operand of a module in movement sparsity.

    This context can resolve the independent structured mask from operand, and can refresh the binary
    mask back to operand with dependent structured mask. Serves as an agent for `StructuredMaskHandler`
    to conduct structured mask resolution.
    """

    def __init__(
        self,
        sparsifier_operand: MovementSparsifier,
        module_node_name: NNCFNodeName,
        grid_size: Tuple[int, int],
        prune_by_row: bool,
    ):
        """
        Initializes the context of the target module for structured masking.

        :param sparsifier_operand: Operand for the target module.
        :param module_node_name: Node name of the target module.
        :param grid_size: The grid shape for resolving the independent structured mask.
        :param prune_by_row: Determines whether to resolve the independent structured mask by row or column.
        """
        self.sparsifier_operand = sparsifier_operand
        self.module_node_name = module_node_name
        operand_mask: torch.Tensor = sparsifier_operand.weight_ctx.binary_mask
        self.operand_mask_shape = operand_mask.shape
        self.grid_size = self._resolve_grid_size(grid_size)
        self.structured_mask_shape = torch.Size(
            dim // grid for dim, grid in zip(self.operand_mask_shape, self.grid_size)
        )
        self.prune_by_row = prune_by_row
        self._independent_structured_mask = None
        self._dependent_structured_mask = None

    def __str__(self) -> str:
        prune_info = "row prune" if self.prune_by_row else "column prune"
        return f'{self.__class__.__name__}({prune_info} by {self.grid_size}, "{self.module_node_name}")'

    @property
    def independent_structured_mask(self) -> Optional[torch.Tensor]:
        if self._independent_structured_mask is None:
            nncf_logger.debug("Independent structured mask has not been calculated. Return None.")
        return self._independent_structured_mask

    @independent_structured_mask.setter
    @torch.no_grad()
    def independent_structured_mask(self, tensor: torch.Tensor):
        if self.structured_mask_shape != tensor.shape:
            raise ValueError("Wrong shape about independent structured mask.")
        if self._independent_structured_mask is None:
            self._independent_structured_mask = tensor.clone()
        else:
            if self._independent_structured_mask.device != tensor.device:
                nncf_logger.debug(f"Changing independent_structured_mask device to {tensor.device}")
                self._independent_structured_mask = self._independent_structured_mask.to(tensor.device)
            self._independent_structured_mask.copy_(tensor)

    @property
    def dependent_structured_mask(self) -> Optional[torch.Tensor]:
        if self._dependent_structured_mask is None:
            nncf_logger.debug("Dependent structured mask has not been calculated. Return None.")
        return self._dependent_structured_mask

    @dependent_structured_mask.setter
    @torch.no_grad()
    def dependent_structured_mask(self, tensor: torch.Tensor):
        if self.structured_mask_shape != tensor.shape:
            raise ValueError("Wrong shape about dependent structured mask.")
        if self._dependent_structured_mask is None:
            self._dependent_structured_mask = tensor.clone()
        else:
            if self._dependent_structured_mask.device != tensor.device:
                nncf_logger.debug(
                    f"Changing dependent_structured_mask device to {tensor.device}",
                )
                self._dependent_structured_mask = self._dependent_structured_mask.to(tensor.device)
            self._dependent_structured_mask.copy_(tensor)

    @torch.no_grad()
    def update_independent_structured_mask_from_operand(self):
        """
        Gets the current unstructured binary mask from operand, resolves it to the independent structured one, and
        stores in `self.independent_structured_mask` for later use in `StructuredMaskHandler`.
        """
        weight_binary_mask = self.sparsifier_operand.weight_ctx.binary_mask.detach().clone()
        mask_by_grid = F.max_pool2d(
            weight_binary_mask.unsqueeze(0), kernel_size=self.grid_size, stride=self.grid_size
        ).squeeze(0)
        preserved_cols = mask_by_grid.amax(dim=0)
        preserved_rows = mask_by_grid.amax(dim=1)

        if self.sparsifier_operand.prune_bias:
            bias_binary_mask = self.sparsifier_operand.bias_ctx.binary_mask.detach().clone()
            bias_preserved_rows = F.max_pool1d(
                bias_binary_mask.view(1, -1), kernel_size=self.grid_size[0], stride=self.grid_size[0]
            ).squeeze(0)
            preserved_rows = bias_preserved_rows.logical_or(preserved_rows)

        structured_mask = preserved_rows.unsqueeze(1) * preserved_cols.unsqueeze(0)
        self.independent_structured_mask = structured_mask
        return structured_mask

    def initialize_binary_mask(self):
        """
        Initialize binary mask by all ones. The inflated dependent mask will be applied via logical "and"
        operation to it. It's needed for the case when 1 binary mask shared for 2 groups: in one group operator
        can be pruned by input channels, i.e. be a consumer of pruning masks, and for another - can be pruned by
        output channels, i.e. be a producer of pruning masks.
        Initial  |  Mask 2 last input channels   |  Mask middle output channel    |     Result
        --------------------------------------------------------------------------------------
         1111                 1100                             1111                     1100
         1111    &            1100               &             0000                =    0000
         1111                 1100                             1111                     1100
        """

        self.sparsifier_operand.weight_ctx.binary_mask.fill_(1)
        if self.sparsifier_operand.prune_bias:
            self.sparsifier_operand.bias_ctx.binary_mask.fill_(1)

    def populate_dependent_structured_mask_to_operand(self):
        """
        Updates the actual binary masks in operand with `self.dependent_structured_mask`.
        """
        structured_mask_inflated = self._inflate_structured_mask(self.dependent_structured_mask, self.grid_size)
        self.sparsifier_operand.weight_ctx.binary_mask *= structured_mask_inflated
        if self.sparsifier_operand.prune_bias:
            self.sparsifier_operand.bias_ctx.binary_mask *= structured_mask_inflated.amax(dim=1)

    def gather_statistics_from_operand(self) -> StructuredMaskContextStatistics:
        """
        Collects the structured mask statistics from the binary masks in operand.

        :return: The statistics of the structured mask context.
        """
        node = self.sparsifier_operand.target_module_node
        assert isinstance(node.layer_attributes, tuple(EXPECTED_NODE_LAYER_ATTRS))
        weight_shape: Tuple[int, int] = tuple(node.layer_attributes.get_weight_shape())
        bias_shape: Tuple[int] = (
            (node.layer_attributes.get_bias_shape(),) if self.sparsifier_operand.prune_bias else (0,)
        )

        pruned_weight_shape = list(weight_shape)
        head_id_to_keep = []
        if self.prune_by_row:
            prunable_rows = self.sparsifier_operand.weight_ctx.binary_mask.amax(dim=1)
            pruned_weight_shape[0] = int(prunable_rows.count_nonzero().item())
            kept_row_blocks = F.max_pool1d(prunable_rows.unsqueeze(0), kernel_size=self.grid_size[0]).squeeze(0)
            head_id_to_keep = kept_row_blocks.nonzero().view(-1).cpu().numpy().tolist()
        else:
            prunable_cols = self.sparsifier_operand.weight_ctx.binary_mask.amax(dim=0)
            pruned_weight_shape[1] = int(prunable_cols.count_nonzero().item())
            kept_col_blocks = F.max_pool1d(prunable_cols.unsqueeze(0), kernel_size=self.grid_size[1]).squeeze(0)
            head_id_to_keep = kept_col_blocks.nonzero().view(-1).cpu().numpy().tolist()

        pruned_bias_shape = bias_shape
        if self.sparsifier_operand.prune_bias and self.prune_by_row:
            pruned_bias_shape = (int(self.sparsifier_operand.bias_ctx.binary_mask.count_nonzero().item()),)

        return StructuredMaskContextStatistics(
            weight_shape=weight_shape,
            pruned_weight_shape=tuple(pruned_weight_shape),
            bias_shape=bias_shape,
            pruned_bias_shape=pruned_bias_shape,
            head_or_channel_id_to_keep=head_id_to_keep,
            module_node_name=self.module_node_name,
        )

    def _resolve_grid_size(self, grid_size) -> Tuple[int, int]:
        a, b = grid_size
        return (a if a > 0 else self.operand_mask_shape[0], b if b > 0 else self.operand_mask_shape[1])

    @staticmethod
    def _inflate_structured_mask(structured_mask: torch.Tensor, grid_size: Tuple[int, int]) -> torch.Tensor:
        assert len(structured_mask.shape) == len(
            grid_size
        ), f"Unmatched dimension with structured_mask in shape {structured_mask.shape} and grid_size in 2D."
        inflated_mask = structured_mask.clone()
        for axis, repeat_times in enumerate(grid_size):
            inflated_mask = inflated_mask.repeat_interleave(repeat_times, dim=axis)
        return inflated_mask


class StructuredMaskContextGroup:
    """
    Stores together the structured mask contexts that are related to the same building block.
    """

    def __init__(self, group_id: int, structured_mask_contexts: List[StructuredMaskContext]):
        """
        Initializes a group of related structured mask contexts.

        :param group_id: The index of the building block.
        :param structured_mask_contexts: A list of structured mask contexts corresponding
            to the building block.
        """
        self.group_id = group_id
        self.structured_mask_contexts = structured_mask_contexts

    def __str__(self) -> str:
        if not self.structured_mask_contexts:
            ctxes_str = "[]"
        else:
            ctxes = (f"\n\t{ctx}" for ctx in self.structured_mask_contexts)
            ctxes_str = "[{}\n]".format("".join(ctxes))
        return f"{self.__class__.__name__}[{self.group_id}]: {ctxes_str}"


class StructuredMaskHandler:
    """
    Handler to conduct structured masking on supported models.

    This handler gathers sparsifiable layers together as groups according to the building block
    they belong to, e.g., multi-head self-attention or feed-forward network in Transformers.
    Within each group, it refreshes the binary masks from unstructured to structured ones,
    while considering the pruning dependencies across layers. All these operations are conducted
    via the `StructuredMaskContext` of each module that supports structured masking.
    """

    def __init__(self, compressed_model: NNCFNetwork, sparsified_module_info_list: List[SparseModuleInfo]):
        """
        Initializes the handler for structured masking in movement sparsity.

        :param compressed_model: The wrapped compressed model.
        :param sparsified_module_info_list: List of `SparsifiedModuleInfo` in the
            controller of `compressed_model`.
        :param strategy: Strategy of resolving structured masks for `compressed_model`.
        """
        self.compressed_model = compressed_model
        self.sparsified_module_info_list = sparsified_module_info_list
        self._structured_mask_ctx_groups = self._create_structured_mask_context_groups(
            compressed_model, sparsified_module_info_list
        )

        nncf_logger.debug("Totally %d structured mask context groups.", len(self._structured_mask_ctx_groups))
        for structured_mask_ctx_group in self._structured_mask_ctx_groups:
            nncf_logger.debug(f"{structured_mask_ctx_group}")

    def update_independent_structured_mask(self):
        """
        Asks all contexts in `self._structured_mask_ctx_groups` to calculate the independent structured mask.
        """
        for group in self._structured_mask_ctx_groups:
            for ctx in group.structured_mask_contexts:
                ctx.update_independent_structured_mask_from_operand()

    def resolve_dependent_structured_mask(self):
        """
        Within each context group, it reads the independent structured masks of related layers and
        resolves the structured masks based on dependency rules defined in `self.rules_by_group_type`.
        """
        for group in self._structured_mask_ctx_groups:
            ctxes = group.structured_mask_contexts
            row_prune_ctxes = list(filter(lambda ctx: ctx.prune_by_row, ctxes))
            col_prune_ctxes = list(filter(lambda ctx: not ctx.prune_by_row, ctxes))
            independent_masks = [ctx.independent_structured_mask for ctx in row_prune_ctxes] + [
                ctx.independent_structured_mask.t() for ctx in col_prune_ctxes
            ]
            coarse_mask = reduce(torch.logical_or, independent_masks).float()
            with torch.no_grad():
                for ctx in row_prune_ctxes:
                    ctx.dependent_structured_mask = coarse_mask
                for ctx in col_prune_ctxes:
                    ctx.dependent_structured_mask = coarse_mask.t()

    def populate_dependent_structured_mask_to_operand(self):
        """
        Asks all contexts in `self._structured_mask_ctx_groups` to update the actual binary masks in operand.
        """
        for group in self._structured_mask_ctx_groups:
            for ctx in group.structured_mask_contexts:
                ctx.initialize_binary_mask()
        for group in self._structured_mask_ctx_groups:
            for ctx in group.structured_mask_contexts:
                ctx.populate_dependent_structured_mask_to_operand()

    def report_structured_sparsity(
        self,
        save_dir: str,
        file_name: str = "structured_sparsity",
        to_csv: bool = True,
        max_num_of_kept_heads_to_report: int = 20,
    ) -> pd.DataFrame:
        """
        Generates a report file that describes the structured mask statistics for each context group.

        :param save_dir: The folder to save the report file.
        :param file_name: File name of the report.
        :param to_csv: Whether to dump the report file in csv format.
        :param max_num_of_kept_heads_to_report: The max number of heads or channels to display that are
            preserved after structured masking. Used to avoid showing too many elements in the list.
        :return: The structured mask statistics in `pandas.DataFrame` format.
        """
        df = self._gather_statistics_dataframe(max_num_of_kept_heads_to_report)
        if to_csv:
            df.to_csv(Path(save_dir, f"{file_name}.csv"))
        return df

    def _gather_statistics_dataframe(self, max_num_of_kept_heads_to_report: int = 20) -> pd.DataFrame:
        module_vs_name_map = {module: name for name, module in self.compressed_model.named_modules()}
        entry_list = []
        for group in self._structured_mask_ctx_groups:
            ctxes = sorted(
                group.structured_mask_contexts, key=lambda ctx: ctx.sparsifier_operand.target_module_node.node_id
            )
            for ctx in ctxes:
                stats = ctx.gather_statistics_from_operand()
                module = self.compressed_model.nncf.get_containing_module(stats.module_node_name)
                entry = dict(group_id=group.group_id, torch_module=module_vs_name_map[module], **stats.__dict__)
                if len(stats.head_or_channel_id_to_keep) > max_num_of_kept_heads_to_report:  # avoid long display
                    entry["head_or_channel_id_to_keep"] = f"[{len(stats.head_or_channel_id_to_keep)} items]"
                entry_list.append(entry)
        return pd.DataFrame(entry_list)

    @staticmethod
    def _create_structured_mask_context_groups(
        nncf_network: NNCFNetwork, sparsified_module_info_list: List[SparseModuleInfo]
    ) -> List[StructuredMaskContextGroup]:
        module_vs_sparse_module_info_map = {minfo.module: minfo for minfo in sparsified_module_info_list}

        pruning_producing_types = ["linear"]
        nncf_graph = nncf_network.nncf.get_original_graph()
        pruning_groups = get_pruning_groups(
            nncf_graph, PT_EXPERIMENTAL_PRUNING_OPERATOR_METATYPES, pruning_producing_types
        )
        pruning_groups = select_largest_groups(pruning_groups)
        result = []
        for group_id, group in enumerate(pruning_groups):
            ctxes = []
            block = group.block
            extended_producers = group.producers
            for consumer in group.consumers:
                if consumer.pruning_dimension is not None:
                    extended_producers.add(ProducerInfo(consumer.node_id, consumer.pruning_dimension))
            is_group_matched = True
            prefix_warning = (
                "Automatically found structured pruning group does not match the given unstructured "
                f"pruning structures: \n{group}\n."
            )
            for producer_info in extended_producers:
                nncf_node = nncf_graph.get_node_by_id(producer_info.node_id)
                module = nncf_network.nncf.get_containing_module(nncf_node.node_name)
                if module in module_vs_sparse_module_info_map:
                    # 0 dimension corresponds to row (output channels), 1st dimension - to column (input channels)
                    prune_by_row = not bool(producer_info.pruning_dimension)
                    minfo = module_vs_sparse_module_info_map[module]
                    sparsifier: MovementSparsifier = minfo.operand
                    if sparsifier.sparse_structure == SparseStructure.PER_DIM:
                        sparse_axis = sparsifier.sparse_cfg.sparse_axis
                        if sparse_axis != producer_info.pruning_dimension:
                            nncf_logger.warning(
                                f"{prefix_warning}. Unstructured pruning is defined for "
                                f"{sparse_axis} axis, but structured one - for "
                                f"{producer_info.pruning_dimension} axis."
                            )
                            is_group_matched = False
                            break
                    prune_grid = (block.size, -1) if prune_by_row else (-1, block.size)
                    ctx = StructuredMaskContext(minfo.operand, minfo.module_node_name, prune_grid, prune_by_row)
                    ctxes.append(ctx)
                else:
                    nncf_logger.warning(
                        f"Automatically found structured pruning group does not match the given "
                        f"unstructured sparse structures:\n {group}"
                    )
                    is_group_matched = False
                    break
            if ctxes and is_group_matched:
                result.append(StructuredMaskContextGroup(group_id, ctxes))
        return result



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Any, Tuple, Type, Union

import torch


def __get_supported_torch_return_types() -> Tuple[Type[tuple], ...]:
    """
    Collects types from torch.return_type which can be wrapped/unwrapped by NNCF.
    NNCF can wrap/unwrap only return types that have two attributes, one of them
    should be the `values` attribute.

    :return: List of types from torch.return_type which can be wrapped/unwrapped by NNCF.
    """
    retval = [t for _, t in inspect.getmembers(torch.return_types) if inspect.isclass(t) and hasattr(t, "values")]
    return tuple(t for t in retval if t.n_fields == 2)


_TORCH_RETURN_TYPES = __get_supported_torch_return_types()


def maybe_get_values_from_torch_return_type(tensor: Any) -> torch.Tensor:
    """
    Attempts to unwrap the tensor value from one of torch.return_types instances
    in case torch operation output is wrapped by a torch return_type.

    :param tensor: Torch tensor or torch return type instance to unwrap values from.
    :return: Unwrapped torch tensor.
    """
    if isinstance(tensor, _TORCH_RETURN_TYPES):
        return tensor.values
    return tensor


def maybe_wrap_to_torch_return_type(
    tensor: torch.Tensor, wrapped_input: Union[tuple, torch.Tensor]
) -> Union[tuple, torch.Tensor]:
    """
    Wraps tensor to wrapped_input wrapper in case wrapped_input is wrapped by a torch.return_value container.

    :param tensor: Torch tensor to wrap.
    :param wrapped_tensor: Instance of the tensor before it was unwrapped.
    :return: Wrapped tensor in case wrapped_input is wrapped by a torch.return_value container else the tensor.
    """

    if isinstance(wrapped_input, _TORCH_RETURN_TYPES):
        # We assume that return_type has only two attributes, the first one is `value`.
        # This assumption is checked by `test_unwrap_wrap_torch_return_type`.
        return wrapped_input.__class__((tensor, wrapped_input[1]))
    return tensor



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

import tensorflow as tf

from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import UnknownMetatype
from nncf.tensorflow.graph.metatypes.keras_layers import KERAS_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.tf_ops import TF_OPERATION_METATYPES


def get_keras_layer_metatype(layer: tf.keras.layers.Layer, determine_subtype: bool = True) -> Type[OperatorMetatype]:
    """
    Returns a metatype of the Keras layer.

    The flag 'determine_subtype' specifies which metatype for the layer the subtype
    or main type will be returned.

    For example, you created instance of the depthwise convolution using
    `tf.keras.layers.Conv2D`, then `get_keras_layer_metatype` will return different
    metatypes depending on the determine_subtype flag.

    ```
    input_shape = (4, 28, 28, 3)
    x = tf.random.normal(input_shape)
    layer = tf.keras.layers.Conv2D(3, 3, groups = 3, input_shape=input_shape[1:])
    y = layer(x)

    metatype = get_keras_layer_metatype(layer, determine_subtype = False)
    assert metatype == TFConv2DLayerMetatype
    metatype = get_keras_layer_metatype(layer, determine_subtype = True)
    assert metatype == TFDepthwiseConv2DSubLayerMetatype
    ```

    :param layer: The Keras layer.
    :param determine_subtype: Determines the subtype of the metatype if True and
        returns the primary metatype otherwise.
    :return: A metatype.
    """
    layer_metatype = KERAS_LAYER_METATYPES.get_operator_metatype_by_op_name(layer.__class__.__name__)

    if not determine_subtype:
        return layer_metatype
    subtype = None
    if layer_metatype is not UnknownMetatype:
        subtype = layer_metatype.determine_subtype(layer)
    if subtype is not None:
        return subtype
    return layer_metatype


def get_op_metatype(op_name: str) -> Type[OperatorMetatype]:
    """
    Returns a metatype of the TF operation by operation name.

    :param op_name: TF operation name.
    :return: A metatype.
    """
    return TF_OPERATION_METATYPES.get_operator_metatype_by_op_name(op_name)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Callable, List, Tuple, TypeVar

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.tensor import Tensor

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")


class SmoothQuantAlgoBackend(ABC):
    @property
    @abstractmethod
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        """
        Parameter for backend-specific metatypes for Convolution.

        :return: OperatorMetatype list.
        """

    @property
    @abstractmethod
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        """
        Parameter for backend-specific metatypes for MatMul.

        :return: OperatorMetatype list.
        """

    @property
    @abstractmethod
    def quantize_agnostic_metatypes(self) -> List[OperatorMetatype]:
        """
        Parameter for backend-specific quantize agnostic metatypes.

        :return: List of OperatorMetatype.
        """

    @staticmethod
    @abstractmethod
    def pre_layer_target_type() -> TargetType:
        """
        Returns backend-specific pre layer target type.

        :returns: Backend-specific pre layer target type.
        """

    @staticmethod
    @abstractmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: Port ID of the tensor for the statistics distribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_weights(node: NNCFNode) -> bool:
        """
        Checks whether the node with weights or not.

        :param node: NNCFNode to check.
        :return: Boolean indicating whether the node has weights or not.
        """

    @staticmethod
    @abstractmethod
    def get_activations_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        """
        Returns map with activation & weighted ports.

        :param node: NNCFNode to check.
        :param nncf_graph: NNCFGraph instance.
        :return: Map with the activation & weighted ports.
        """

    @staticmethod
    @abstractmethod
    def get_abs_max_channel_collector(
        num_samples: int, stats_reduction_axes: Tuple[int], inplace: bool, branch_key: str
    ) -> TensorCollector:
        """
        Returns TensorCollector with MaxAggregator and AbsMaxReducer.

        :param stats_reduction_axes: Calculated reduction axes.
        :param inplace: Whether to calculate statistic inplace or not.
        :param branch_key: Specific string for branch key.
        :return: TensorCollector instance.
        """

    @staticmethod
    @abstractmethod
    def get_weight_value(node_with_weight: NNCFNode, model: TModel, port_id: int) -> Tensor:
        """
        Returns the weight value for the node with weight.

        :param node_with_weight: The node with weight.
        :param model: The model that contains this operation.
        :param port_id: The input port ID to get weight input.
        :return: The weight value.
        """

    @staticmethod
    @abstractmethod
    def get_weight_tensor_port_id(node: NNCFNode) -> int:
        """
        Returns node's input port indices with weights tensors.

        :param node: NNCFNode to find its weights input port indices.
        :return: Weights input port indices.
        """

    @staticmethod
    @abstractmethod
    def weight_update_command(
        node_with_weight: NNCFNode, weight_value: TTensor, weight_port_id: int
    ) -> TransformationCommand:
        """
        Returns command to update weights.

        :param node_with_weight: NNCFNode instance.
        :param weight_value: New weight value.
        :param weight_port_id: Weight port id.
        :return: TransformationCommand instance.
        """

    @staticmethod
    @abstractmethod
    def scale_insertion_command(
        source_node: NNCFNode, scale_value: TTensor, source_output_port_id: int, nodes: List[NNCFNode]
    ) -> TransformationCommand:
        """
        Returns command to insert Smooth Quant node.

        :param source_node: NNCFNode instance.
        :param scale_value: Smooth Quant value.
        :param source_output_port_id: Output port for source node.
        :param nodes: List of consumers for Smooth node.
        :return: TransformationCommand instance.
        """

    @staticmethod
    @abstractmethod
    def get_activation_channel_axis(node: NNCFNode, port_id: int) -> int:
        """
        Returns axis number of the activation tensor which correspond to it channel.

        :param node: NNCFNode instance.
        :param port_id: Specified input port id.
        :return: Channel axis number.
        """

    @staticmethod
    @abstractmethod
    def get_weight_channel_axis(node: NNCFNode) -> int:
        """
        Returns axis number of the weight tensor which correspond to it channel.

        :param node: NNCFNode instance.
        :return: Channel axis number.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_shared_weight(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Returns true if given node shares constant with a different node.

        :param node: NNCFNode instance.
        :param nncf_graph: NNCFGraph instance.
        :return: Whether the given node is shares weights with a different node or not.
        """

    @staticmethod
    @abstractmethod
    def get_filter_fn_for_statistics(activation_port_id: int, algorithm_key: str) -> Callable[[StatisticPoint], bool]:
        """
        Returns backend-specific callable to filter statistic containers according to its statistic point.

        :param activation_port_id: Activation port id for the statistic collection target node.
        :param algorithm_key: Current algorithm key.
        :return: Backend-specific callable to filter statistic containers according to its statistic point.
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NoReturn, Optional, Tuple, TypeVar

from nncf.common.logging import nncf_logger

BaseEvaluatorType = TypeVar("BaseEvaluatorType")
ElasticControllerType = TypeVar("ElasticControllerType")
SearchParametersType = TypeVar("SearchParametersType")


class BaseEvaluatorHandler:
    """
    An interface for handling an evaluator. Evaluator handlers initialize the underlying
    evaluator and can be used to update the evaluator's properties.

    """

    def __init__(self, evaluator: BaseEvaluatorType, elasticity_ctr: ElasticControllerType):
        """
        Initializes the evaluator handler

        :param evaluator: An interface for collecting measurements at a target device.
        :param elasticity_ctr: interface to manage the elasticity of the super-network.
        """
        self.evaluator = evaluator
        self.elasticity_ctrl = elasticity_ctr
        self.elasticity_ctrl.multi_elasticity_handler.activate_maximum_subnet()
        self.input_model_value = evaluator.evaluate_subnet()

    @property
    def name(self):
        return self.evaluator.name

    @property
    def current_value(self):
        return self.evaluator.current_value

    def retrieve_from_cache(self, subnet_config_repr: Tuple[float, ...]) -> Tuple[bool, float]:
        return self.evaluator.retrieve_from_cache(subnet_config_repr)

    def evaluate_and_add_to_cache_from_pymoo(self, pymoo_repr: Tuple[float, ...]) -> float:
        return self.evaluator.evaluate_and_add_to_cache_from_pymoo(pymoo_repr)

    def export_cache_to_csv(self, cache_file_path: str) -> NoReturn:
        self.evaluator.export_cache_to_csv(cache_file_path)


class EfficiencyEvaluatorHandler(BaseEvaluatorHandler):
    """
    An interface for handling efficiency evaluators
    """


class AccuracyEvaluatorHandler(BaseEvaluatorHandler):
    """
    An interface for handling accuracy evaluators

    """

    def __init__(self, accuracy_evaluator, elasticity_ctrl, ref_acc: Optional[float] = 100):
        super().__init__(accuracy_evaluator, elasticity_ctrl)
        self._ref_acc = ref_acc

    @property
    def ref_acc(self) -> float:
        """
        :return: reference accuracy
        """
        return self._ref_acc

    @ref_acc.setter
    def ref_acc(self, val: float) -> NoReturn:
        """
        :param val: value to update the reference accuracy value.
        :return:
        """
        self._ref_acc = val

    def update_reference_accuracy(self, search_params: SearchParametersType) -> NoReturn:
        """
        Update reference accuracy of the search algorithm

        :param search_params: parameters of the search algorithm
        :return:
        """
        self.ref_acc = search_params.ref_acc
        if self.input_model_value > self.ref_acc - 0.01 or self.input_model_value < self.ref_acc + 0.01:
            nncf_logger.warning(
                f"Accuracy obtained from evaluation {self.input_model_value} "
                f"differs from reference accuracy {self.ref_acc}"
            )
            if self.ref_acc == -1:
                nncf_logger.info("Adjusting reference accuracy to accuracy obtained from evaluation")
                self.ref_acc = self.input_model_value
            else:
                if self.ref_acc >= 100:
                    nncf_logger.error(f"Reference accuracy value is invalid: {self.ref_acc}")
                nncf_logger.info("Using reference accuracy.")
                self.input_model_value = self.ref_acc
        search_params.ref_acc = self.ref_acc



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.openvino.graph.metatypes.groups import FAKE_QUANTIZE_OPERATIONS
from nncf.openvino.graph.model_utils import remove_fq_from_inputs
from nncf.openvino.graph.node_utils import get_bias_value
from nncf.openvino.graph.node_utils import get_parameter_node_name
from nncf.openvino.graph.node_utils import get_result_node_name
from nncf.openvino.graph.node_utils import is_node_with_bias
from nncf.openvino.graph.transformations.command_creation import OVCommandCreator
from nncf.openvino.graph.transformations.commands import OVBiasCorrectionCommand
from nncf.openvino.graph.transformations.commands import OVModelExtractionCommand
from nncf.openvino.graph.transformations.commands import OVOutputInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.statistics.collectors import OVNNCFCollectorTensorProcessor
from nncf.openvino.statistics.collectors import get_mean_statistic_collector
from nncf.openvino.statistics.collectors import get_raw_stat_collector
from nncf.openvino.tensor import OVNNCFTensor
from nncf.quantization.algorithms.bias_correction.backend import BiasCorrectionAlgoBackend


class OVBiasCorrectionAlgoBackend(BiasCorrectionAlgoBackend):
    @property
    def tensor_processor(self) -> OVNNCFCollectorTensorProcessor:
        return OVNNCFCollectorTensorProcessor

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: np.ndarray, nncf_graph: NNCFGraph
    ) -> OVBiasCorrectionCommand:
        return OVCommandCreator.create_command_to_update_bias(node, bias_value, nncf_graph)

    @staticmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> OVModelExtractionCommand:
        return OVModelExtractionCommand(input_ids, output_ids)

    @staticmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: OVTargetPoint) -> OVOutputInsertionCommand:
        return OVOutputInsertionCommand(target_point)

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size, inplace)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: str) -> OVNNCFTensor:
        return OVNNCFTensor(raw_data[output_name])

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        constant_ports = node.layer_attributes.get_const_port_ids()
        activation_ports = [
            e.input_port_id for e in nncf_graph.get_input_edges(node) if e.input_port_id not in constant_ports
        ]
        assert len(activation_ports) == 1
        return activation_ports[0]

    @staticmethod
    def get_bias_value(node: NNCFNode, model: ov.Model, nncf_graph: NNCFGraph) -> np.ndarray:
        return get_bias_value(node, nncf_graph, model)

    @staticmethod
    def get_input_name(model: ov.Model, node_name: str, input_port_id: int) -> str:
        return get_parameter_node_name(node_name, input_port_id)

    @staticmethod
    def get_output_name(model: ov.Model, node_name: str, output_port_id: int) -> str:
        return get_result_node_name(node_name, output_port_id)

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        if node.layer_attributes is None:
            return False
        const_port_ids = node.layer_attributes.get_const_port_ids()
        assert len(const_port_ids) == 1
        weight_node = nncf_graph.get_input_edges(node)[const_port_ids[0]].from_node
        return weight_node.metatype in FAKE_QUANTIZE_OPERATIONS

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def remove_fq_from_inputs(model: ov.Model, nncf_graph: NNCFGraph) -> ov.Model:
        return remove_fq_from_inputs(model, nncf_graph)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from contextlib import contextmanager

NNCF_LOGGER_NAME = "nncf"

nncf_logger = logging.getLogger(NNCF_LOGGER_NAME)
nncf_logger.propagate = False

stdout_handler = logging.StreamHandler(sys.stdout)
fmt_string = "%(levelname)s:%(name)s:%(message)s"
fmt = logging.Formatter(fmt_string)
fmt_no_newline = logging.Formatter(
    fmt_string,
)
stdout_handler.setFormatter(fmt)
stdout_handler.setLevel(logging.INFO)
nncf_logger.addHandler(stdout_handler)
nncf_logger.setLevel(logging.INFO)


def set_log_level(level: int):
    """
    Sets the log level for the NNCF logging.
    :param level: An integer passed into the underlying `logging.Logger` object,
      i.e. the regular `logging` log levels must be used here such as `logging.WARNING`
      or `logging.DEBUG`.
    """
    nncf_logger.setLevel(level)
    for handler in nncf_logger.handlers:
        handler.setLevel(level)


def set_log_file(filename: str):
    """
    Sets the log file for the NNCF logging.

    :param filename: Path to the file to save the log.
    """
    file_handler = logging.FileHandler(filename, encoding="utf-8")
    file_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
    nncf_logger.addHandler(file_handler)


def disable_logging():
    """
    Disables NNCF logging entirely. `FutureWarning`s are still shown.
    """
    nncf_logger.handlers = []


class DuplicateFilter:
    def __init__(self):
        self.msgs = set()

    def filter(self, rec):
        retval = rec.msg not in self.msgs
        self.msgs.add(rec.msg)
        return retval


NNCFDeprecationWarning = FutureWarning


@contextmanager
def extension_is_loading_info_log(extension_name: str):
    nncf_logger.info(f"Compiling and loading torch extension: {extension_name}...")
    yield
    nncf_logger.info(f"Finished loading torch extension: {extension_name}")


def warn_bkc_version_mismatch(backend: str, bkc_version: str, current_version: str):
    nncf_logger.warning(
        f"NNCF provides best results with {backend}{bkc_version}, "
        f"while current {backend} version is {current_version}. "
        f"If you encounter issues, consider switching to {backend}{bkc_version}"
    )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional, Tuple, Union


class Dtype(Enum):
    FLOAT = "float"
    INTEGER = "int"


class BaseLayerAttributes(ABC):
    """
    This class stores base useful for some algorithms attributes
    of modules/layers.
    """

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, self.__class__) and self.__dict__ == __o.__dict__


class MultipleInputLayerAttributes(BaseLayerAttributes):
    def __init__(self, axis: int, num_inputs: Optional[int] = None):
        """

        :param axis: the dimension over which the inputs are combined (e.g. concatenated).
        :param num_inputs: Number of inputs.
        """
        self.axis = axis
        self.num_inputs = num_inputs


class MultipleOutputLayerAttributes(BaseLayerAttributes):
    def __init__(self, chunks: Union[int, List[Any]], axis: int):
        """

        :param chunks: Number of chunks (outputs).
        :param axis: The dimension along which to make multiple outputs (e.g. split the tensor).
        """
        self.chunks = chunks
        self.axis = axis


class WeightedLayerAttributes(BaseLayerAttributes):
    def __init__(self, weight_requires_grad: bool, dtype: Dtype = Dtype.FLOAT, with_bias: bool = False):
        """

        :param weight_requires_grad: Is True if gradients need to be computed for the corresponding Tensor,
        False otherwise.
        :param dtype: is an object that represents the type of data.
        :param with_bias: Operation include bias.
        """
        self.weight_requires_grad = weight_requires_grad
        self.dtype = dtype
        self.with_bias = with_bias

    @abstractmethod
    def get_weight_shape(self) -> List[int]:
        pass

    def get_num_filters(self) -> int:
        weight_shape = self.get_weight_shape()
        return weight_shape[self.get_target_dim_for_compression()]

    @abstractmethod
    def get_target_dim_for_compression(self) -> int:
        pass


class GenericWeightedLayerAttributes(WeightedLayerAttributes):
    """
    Represents a weighted layer for which there is no information ahead of time
    of the exact meaning of the weight indices.
    """

    def __init__(
        self,
        weight_requires_grad: bool,
        weight_shape: List[int],
        filter_dimension_idx: int = 0,
        with_bias: bool = False,
    ):
        """

        :param weight_requires_grad: Is True if gradients need to be computed for the corresponding Tensor,
        False otherwise.
        :param weight_shape: shape of weight tensor.
        :param filter_dimension_idx: the axis, along which the filters are stored.
        """
        super().__init__(weight_requires_grad=weight_requires_grad, with_bias=with_bias)
        self.weight_shape = weight_shape
        self.filter_dimension_idx = filter_dimension_idx

    def get_weight_shape(self) -> List[int]:
        return self.weight_shape

    def get_target_dim_for_compression(self) -> int:
        return 0


class LinearLayerAttributes(WeightedLayerAttributes):
    def __init__(
        self,
        weight_requires_grad: bool,
        in_features: int,
        out_features: int,
        with_bias: bool = True,
    ):
        """

        :param weight_requires_grad: Is True if gradients need to be computed for the corresponding Tensor,
        False otherwise.
        :param in_features: number of input channels in the layer's input.
        :param out_features: number of channels produced by the layer.
        """
        super().__init__(weight_requires_grad, with_bias=with_bias)
        self.in_features = in_features
        self.out_features = out_features

    def get_weight_shape(self) -> List[int]:
        return [self.out_features, self.in_features]

    def get_bias_shape(self) -> int:
        return self.out_features if self.with_bias is True else 0

    def get_target_dim_for_compression(self) -> int:
        return 0


class ConvolutionLayerAttributes(WeightedLayerAttributes):
    def __init__(
        self,
        weight_requires_grad: bool,
        in_channels: int,
        out_channels: int,
        kernel_size: Tuple[int, ...],
        stride: Tuple[int, ...],
        dilations: Tuple[int, ...],
        groups: int,
        transpose: bool,
        padding_values: Union[Tuple[int, ...], int],
        with_bias: bool = False,
        output_padding_values: Optional[Union[Tuple[int, ...], int]] = None,
    ):
        """

        :param weight_requires_grad: Is True if gradients need to be computed for the corresponding Tensor,
        False otherwise.
        :param in_channels: Number of input channels in the layer's input.
        :param out_channels: Number of channels produced by the layer.
        :param kernel_size: Size of the convolving kernel.
        :param stride: Stride of the convolution.
        :param groups: Number of blocked connections from input channels to output channels.
        :param transpose: If set to `True`, the layer is an ordinary convolution, otherwise - transpose one.
        :param padding_values: Defines the amount of padding applied to the layer's input.
        :param with_bias: Operation include bias.
        :param output_padding_values: Defines the amount of output padding applied to the layer's output, for transpose.
        """
        super().__init__(weight_requires_grad=weight_requires_grad, with_bias=with_bias)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilations = dilations
        self.groups = groups
        self.transpose = transpose
        self.padding_values = padding_values
        self.output_padding_values = output_padding_values

    def get_weight_shape(self) -> List[int]:
        if not self.transpose:
            return [self.out_channels, self.in_channels // self.groups, *self.kernel_size]
        return [self.in_channels, self.out_channels // self.groups, *self.kernel_size]

    def get_target_dim_for_compression(self) -> int:
        # Always quantize per each "out" channel
        if self.transpose:
            return 1
        return 0


class GroupNormLayerAttributes(WeightedLayerAttributes):
    def __init__(self, weight_requires_grad: bool, num_channels: int, num_groups: int):
        """

        :param weight_requires_grad: Is True if gradients need to be computed for the corresponding Tensor,
        False otherwise.
        :param num_channels: number of channels expected in the layer's input.
        :param num_groups: number of groups to separate the channels into.
        """
        super().__init__(weight_requires_grad)
        self.num_channels = num_channels
        self.num_groups = num_groups

    def get_weight_shape(self) -> List[int]:
        return [self.num_channels]

    def get_target_dim_for_compression(self) -> int:
        return 0


@dataclass
class ReshapeLayerAttributes(BaseLayerAttributes):
    """
    :param input_shape: number of elements of each of the axes of a input tensor.
    :param output_shape: number of elements of each of the axes of a output tensor.
    """

    input_shape: List[int]
    output_shape: List[int]


@dataclass
class TransposeLayerAttributes(BaseLayerAttributes):
    """
    :param dim0: the first dimension to be transposed.
    :param dim1: the second dimension to be transposed.
    """

    dim0: int
    dim1: int


@dataclass
class PermuteLayerAttributes(BaseLayerAttributes):
    """
    :param permutation: the desired ordering of dimensions.
    """

    permutation: Tuple[int, ...]


@dataclass
class GetItemLayerAttributes(BaseLayerAttributes):
    """
    :param key: usually int, tuple of int or slice.
    """

    key: Any


@dataclass
class PadLayerAttributes(BaseLayerAttributes):
    """
    :param mode: mode of the padding operation.
    :param value: fill value of the padding operation.
    """

    mode: str = "constant"
    value: float = 0


@dataclass
class ConvertDtypeLayerAttributes(BaseLayerAttributes):
    """
    :param src_dtype: node input data type.
    :param dst_dtype: node output data type.
    """

    src_dtype: Any
    dst_dtype: Any


@dataclass
class ConstantLayerAttributes(BaseLayerAttributes):
    """
    :param name: Constant name.
    :param shape: Constant shape.
    """

    name: str
    shape: List[int]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Optional

import tensorflow as tf

_CURRENT_CONTEXT = threading.local()


def get_current_context():
    """
    Returns current active `TFTracingContext`.

    :return: Tracing context.
    """
    tracing_context = getattr(_CURRENT_CONTEXT, "tracing_context", None)
    if tracing_context is None:
        tracing_context = TFTracingContext()
        setattr(_CURRENT_CONTEXT, "tracing_context", tracing_context)
    return tracing_context


class TFTracingContextState:
    """
    Contains values that describe a state of the `TFTracingContext`.
    """

    def __init__(self, in_call: bool = False, wrap_ops: bool = False, model: Optional[tf.keras.Model] = None):
        """
        Initializes the `TFTracingContextState` instance.

        :param in_call: Whether currently inside the `call()` method of a `tf.keras.Model`.
        :param wrap_ops: Whether currently adding the compression pre-hooks and post-hooks
            to TensorFlow operations.
        :param model: The Keras model whose `call()` method is currently active.
            The `None` value is specified that model is undefined at this moment. This is only
            possible when `in_call` is equal to `False`.
        """
        self._in_call = in_call
        self._wrap_ops = wrap_ops

        if model is None and in_call:
            raise ValueError(
                f"Inconsisten values `{in_call}` and `{model}` for `in_call` and `model` parameters. "
                "The `None` value is specified that model is undefined at this moment. This is only "
                "possible when `in_call` is equal to `False`."
            )

        self._model = model

    @property
    def in_call(self) -> bool:
        return self._in_call

    @property
    def wrap_ops(self) -> bool:
        return self._wrap_ops

    @property
    def model(self) -> tf.keras.Model:
        return self._model


class TFTracingContext:
    """
    Contains information about should we wrap the TensorFlow
    operation or not.
    """

    def __init__(self):
        """
        Initializes the `TFTracingContext` instance.
        """
        self._state = TFTracingContextState()
        # Maps a name used in the graph to the next id to use for that name.
        self.names_in_use = {}

    @property
    def model(self) -> Optional[tf.keras.Model]:
        return self.state.model

    @property
    def in_call(self) -> bool:
        return self.state.in_call

    @property
    def wrap_ops(self) -> bool:
        return self.state.wrap_ops

    def enter(self, in_call: bool, wrap_ops: bool, model: Optional[tf.keras.Model] = None):
        """
        Pushes parameters onto the tracing context.

        :param in_call: Whether currently inside the `call()` method of a model.
        :param wrap_ops: Whether currently adding the compression pre-hooks and post-hooks
            to TensorFlow operations.
        :param model: The Keras model whose `call()` method is currently active.
        """
        model = self.state.model if model is None else model
        next_state = TFTracingContextState(in_call, wrap_ops, model)
        return TFTracingContextManager(self, next_state)

    def unique_name(self, name: str) -> str:
        """
        Returns a unique operation name for `name`.

        For more details, please, see implementation of
        the `tf.Graph.unique_name()` method.

        :param name: The name for an operation.
        :return: Unique name.
        """
        name_key = name.lower()

        i = self.names_in_use.get(name_key, 0)
        self.names_in_use[name_key] = i + 1

        if i > 0:
            base_name_key = name_key
            # Make sure the composed name key is not already used.
            while name_key in self.names_in_use:
                name_key = f"{base_name_key}_{i}"
                i += 1

            # Mark the composed name_key as used in case someone wants
            # to call unique_name('name_1').
            self.names_in_use[name_key] = 1

            # Return the new name with the original capitalization of the given name.
            name = f"{name}_{i - 1}"
        return name

    @property
    def state(self) -> TFTracingContextState:
        return self._state

    def load_state(self, state: TFTracingContextState) -> None:
        self._state = state


class TFTracingContextManager:
    """
    Context manager for the tracing context.
    """

    def __init__(self, tracing_context: TFTracingContext, next_state: TFTracingContextState):
        """
        Initializes the tracing context manager.

        :param tracing_context: Tracing context.
        :param next_state: Next state of the tracing context which
            should be applied.
        """
        self._tracing_context = tracing_context
        self._next_state = next_state
        self._prev_state = None

    def __enter__(self):
        self._prev_state = self._tracing_context.state
        self._tracing_context.load_state(self._next_state)

    def __exit__(self, *exc):
        self._tracing_context.load_state(self._prev_state)

        if not self._tracing_context.in_call:
            self._tracing_context.names_in_use = {}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

TARGET_DEVICE = "ANY"

NUM_BN_ADAPTATION_SAMPLES = 2000
NUM_INIT_SAMPLES = 256
MIN_PERCENTILE = 0.1
MAX_PERCENTILE = 99.9

PRECISION_INIT_BITWIDTHS = [2, 4, 8]
HAWQ_NUM_DATA_POINTS = 100
HAWQ_ITER_NUMBER = 200
HAWQ_TOLERANCE = 1e-4
HAWQ_COMPRESSION_RATIO = 1.5
HAWQ_DUMP_INIT_PRECISION_DATA = False

AUTOQ_ITER_NUMBER = 0
AUTOQ_COMPRESSION_RATIO = 0.15  # TODO (vshampor) do autoq and hawq follow the same convention here?
AUTOQ_EVAL_SUBSET_RATIO = 1.0
AUTOQ_WARMUP_ITER_NUMBER = 20

QUANTIZATION_PRESET = "performance"
QUANTIZE_INPUTS = True
QUANTIZE_OUTPUTS = False
QUANTIZATION_EXPORT_TO_ONNX_STANDARD_OPS = False
QUANTIZATION_OVERFLOW_FIX = "enable"
QUANTIZATION_BITS = 8
QUANTIZATION_PER_CHANNEL = False
QUANTIZATION_LOGARITHM_SCALE = False

ACTIVATIONS_QUANT_START_EPOCH = 1
WEIGHTS_QUANT_START_EPOCH = 1
LR_POLY_DURATION_EPOCHS = 30
STAGED_QUANTIZATION_BASE_LR = 1e-3
STAGED_QUANTIZATION_BASE_WD = 1e-5

PRUNING_INIT = 0.0
PRUNING_SCHEDULE = "exponential"
PRUNING_TARGET = 0.5
PRUNING_NUM_INIT_STEPS = 0
PRUNING_STEPS = 100
PRUNING_FILTER_IMPORTANCE = "L2"
PRUNING_INTERLAYER_RANKING_TYPE = "unweighted_ranking"
PRUNING_ALL_WEIGHTS = False
PRUNE_FIRST_CONV = False
PRUNE_BATCH_NORMS = True
PRUNE_DOWNSAMPLE_CONVS = False

PRUNING_LEGR_GENERATIONS = 400
PRUNING_LEGR_TRAIN_STEPS = 200
PRUNING_LEGR_MAX_PRUNING = 0.8
PRUNING_LEGR_RANDOM_SEED = 42
PRUNING_LEGR_POPULATION_SIZE = 64
PRUNING_LEGR_NUM_SAMPLES = 16
PRUNING_LEGR_MUTATE_PERCENT = 0.1
PRUNING_LEGR_SIGMA_SCALE = 1

SPARSITY_INIT = 0.0
MAGNITUDE_SPARSITY_WEIGHT_IMPORTANCE = "normed_abs"
SPARSITY_SCHEDULER = "polynomial"
RB_SPARSITY_SCHEDULER = "exponential"
SPARSITY_SCHEDULER_PATIENCE = 1
SPARSITY_SCHEDULER_POWER = 0.9
SPARSITY_SCHEDULER_CONCAVE = True
SPARSITY_TARGET = 0.5
SPARSITY_TARGET_EPOCH = 90
SPARSITY_FREEZE_EPOCH = 100
SPARSITY_SCHEDULER_UPDATE_PER_OPTIMIZER_STEP = False
SPARSITY_MULTISTEP_STEPS = [90]
SPARSITY_MULTISTEP_SPARSITY_LEVELS = [0.1, 0.5]
SPARSITY_LEVEL_SETTING_MODE = "global"

KNOWLEDGE_DISTILLATION_SCALE = 1.0
KNOWLEDGE_DISTILLATION_TEMPERATURE = 1.0

AA_PATIENCE_EPOCHS = 3
AA_INITIAL_COMPRESSION_RATE_STEP = 0.1
AA_INITIAL_TRAINING_PHASE_EPOCHS = 5
AA_COMPRESSION_RATE_STEP_REDUCTION_FACTOR = 0.5
AA_LR_REDUCTION_FACTOR = 0.5
AA_MINIMAL_COMPRESSION_RATE_STEP = 0.025
AA_MAXIMAL_TOTAL_EPOCHS = 10000

VALIDATE_SCOPES = True



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Dict, Iterable, List, Optional, Tuple, TypeVar

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.experimental.tensor import Tensor
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters

TModel = TypeVar("TModel")


class WeightCompressionAlgoBackend(ABC):
    @property
    @abstractmethod
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific metatypes for matmul layers.
        """

    @property
    @abstractmethod
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific metatypes for convolution layers.
        """

    @property
    @abstractmethod
    def embedding_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific metatypes for embedding layers.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_weights(node: NNCFNode, graph: NNCFGraph) -> bool:
        """
        Checks whether the node with weights or not.

        :param node: The node to check.
        :param graph: The model graph.
        :return: True if the node contains weights, False otherwise.
        """

    @staticmethod
    @abstractmethod
    def get_reduction_axes(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Optional[Tuple[int]]:
        """
        Returns reduction axes without axes that corresponds to weight channels of the node with weight.

        :param node_with_weight: The node with weight.
        :param weight_port_id: The input port ID that corresponds to weight.
        :param graph: The model graph.
        :return: Reduction shape in tuple format or None if not applicable.
        """

    @staticmethod
    @abstractmethod
    def get_weight_names_and_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Tuple[str, int]]:
        """
        Returns a list of weight names and port ids for the given node.

        :param node: The node.
        :param graph: The model graph.
        :return: List of tuples with weight names and port ids.
        """

    @abstractmethod
    def get_weight(self, node_with_weight: NNCFNode, weight_port_id: int, model: TModel, graph: NNCFGraph) -> Tensor:
        """
        Returns a weight associated with the given node on the given port id.

        :param node_with_weight: The node with weight.
        :param weight_port_id: The weight port id for given node with weight.
        :param model: The model.
        :param graph: The model graph associated with the model.
        :return: The weight tensor.
        """

    @abstractmethod
    def set_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: TModel, graph: NNCFGraph, weight: Tensor
    ) -> None:
        """
        Update a weight associated with the given node on the given port id.

        :param node_with_weight: The node with weight.
        :param weight_port_id: The weight port id for given node with weight.
        :param model: The model.
        :param graph: The model graph associated with the model.
        :param weight: The weight tensor.
        """

    @abstractmethod
    def transform_model(
        self,
        model: TModel,
        graph: NNCFGraph,
        weight_compression_parameters: Iterable[WeightCompressionParameters],
        precomputed_scales: Dict[str, Tensor] = None,
        precomputed_zero_points: Dict[str, Tensor] = None,
    ) -> TModel:
        """
        Applies weight compression transformations to the model.

        :param model: Model in which the weights will be compressed according to the weight compression description.
        :param graph: The graph associated with the model.
        :param weight_compression_parameters: List of weight compression parameters.
        :param precomputed_scales: Precomputed scales for weights compression.
        :param precomputed_zero_points: Precomputed zero points for weights compression.
        :return: The transformed model.
        """

    @staticmethod
    @abstractmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: id of the port for the statistics distribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorStatisticCollectorBase:
        """
        Returns backend-specific raw statistic collector.
        This statistic collector is used for raw data calculation, without aggregating.

        :param num_samples: Maximum number of samples to collect.
        :return: Backend-specific TensorStatisticCollectorBase for the statistics calculation.
        """

    @staticmethod
    @abstractmethod
    def get_activation_port_id(node: NNCFNode, graph: NNCFGraph) -> int:
        """
        Returns input port id corresponding to activation input edge for the node.
        Supports only nodes that could have bias value.

        :param node: Node of NNCFGraph with bias value.
        :param graph: NNCFGraph instance with the node.
        :return: target input port id.
        """

    @staticmethod
    def dump_parameters(
        model: TModel, parameters: Dict, algo_name: Optional[str] = "quantization", path: Optional[List] = None
    ) -> None:
        """
        Dumps the given parameters into Model's meta section.

        :param model: ov.Model instance.
        :param algo_name: Name of the algorithm to which the parameters refer.
        :param parameters: Incoming dictionary with parameters to save.
        :param path: Optional list of the paths.
        """


class AWQAlgoBackend(WeightCompressionAlgoBackend):
    @staticmethod
    def get_awq_patterns() -> Dict:
        """
        Returns patterns of nodes in network graph for applying AWQ algorithm.
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from collections import OrderedDict
from collections import namedtuple
from typing import Callable, Dict, List, Set, Union

import tensorflow as tf

import nncf
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationType
from nncf.tensorflow.graph.transformations.commands import TFAfterLayer
from nncf.tensorflow.graph.transformations.commands import TFBeforeLayer
from nncf.tensorflow.graph.transformations.commands import TFLayer
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.commands import TFMultiLayerPoint
from nncf.tensorflow.graph.transformations.commands import TFOperationWithWeights
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import get_custom_objects
from nncf.tensorflow.graph.utils import get_weight_name
from nncf.tensorflow.graph.utils import is_functional_model
from nncf.tensorflow.graph.utils import is_sequential_or_functional_model
from nncf.tensorflow.graph.utils import reformat_inbound_nodes_for_oplambda
from nncf.tensorflow.layers.custom_objects import get_nncf_custom_objects
from nncf.tensorflow.layers.wrapper import NNCFWrapper

WeightOperations = namedtuple("WeightOperations", ("weights_attr_name", "operations"))


class TFModelTransformer(ModelTransformer):
    """
    Applies transformations to a Keras model graph.
    """

    def __init__(self, model):
        """
        Initializes Model Transformer

        :param model: Keras model to be transformed
        """
        if not is_sequential_or_functional_model(model):
            raise ValueError("Only tf.keras sequential or functional models can be transformed.")

        super().__init__(model)
        self._model_config = model.get_config()
        self._custom_objects = dict(list(get_custom_objects(model).items()) + list(get_nncf_custom_objects().items()))
        self._name_mapping = {}

    def transform(self, transformation_layout: TFTransformationLayout):
        """Applies transformations to the Keras model.

        :param transformation_layout: List of transformations
        :return: The transformed Keras model
        """
        layer_weights_map = {layer.name: self._get_layer_weights(layer) for layer in self._model.layers}

        for transform in transformation_layout.transformations:
            self._apply_transformation(transform)

        if is_functional_model(self._model):
            transformed_model = tf.keras.Model.from_config(self._model_config, self._custom_objects)
        else:
            transformed_model = tf.keras.Sequential.from_config(self._model_config, self._custom_objects)

        for layer in transformed_model.layers:
            weights = layer_weights_map.get(layer.name)
            if weights:
                self._set_layer_weights(layer, weights)

        return transformed_model

    def _get_layer_weights(self, layer):
        weights_map = OrderedDict()
        for weight_tensor, weight_numpy in zip(layer.weights, layer.get_weights()):
            weights_map[get_weight_name(weight_tensor.name, layer.name)] = weight_numpy

        return weights_map

    @staticmethod
    def _set_layer_weights(layer, weights_map: Dict):
        weight_value_tuples = []
        for weight_tensor in layer.weights:
            weight_name = get_weight_name(weight_tensor.name, layer.name)
            if weight_name in weights_map:
                weight_value_tuples.append((weight_tensor, weights_map[weight_name]))

        tf.keras.backend.batch_set_value(weight_value_tuples)

    def _get_layer(self, layer_name: str):
        # For purposes of performance, will try to get the original unmodified layer
        # first. This function could have been made to return the modified layer state by
        # first trying to deserialize the corresponding name from the config (which is potentially
        # compression-modified at this stage), but then the time required to process e.g. all of the weight
        # insertions in complex custom layers becomes prohibitive since for each weight adjustment in the same layer
        # we would have to deserialize the layer object.
        for layer in self._model.layers:
            if layer.name == layer_name:
                return layer

        _, layer_config = self._find_layer_config(layer_name)
        if layer_config:
            return tf.keras.utils.deserialize_keras_object(layer_config, custom_objects=self._custom_objects)

        return None

    def _find_layer_config(self, layer_name: str):
        for idx, layer in enumerate(self._model_config["layers"]):
            layer_name_ = layer["name"] if is_functional_model(self._model) else layer["config"]["name"]
            if layer_name_ == layer_name:
                return idx, layer
        return None, None

    def _update_layer_mapping(self, src_layer_name: str, dst_layer_name: str):
        if src_layer_name in self._name_mapping.values():
            for orig_layer_name, orig_layer in self._name_mapping.items():
                if orig_layer == src_layer_name:
                    self._name_mapping[orig_layer_name] = dst_layer_name
        else:
            self._name_mapping[src_layer_name] = dst_layer_name

    def _apply_transformation(self, transformation: TransformationCommand):
        if transformation.type == TransformationType.INSERT:
            self._insert(transformation.target_point, transformation.insertion_objects)
        elif transformation.type == TransformationType.MULTI_INSERT:
            self._multi_insertion(transformation.target_point, transformation.commands)
        elif transformation.type == TransformationType.REMOVE:
            self._remove(transformation.target_point)
        else:
            raise TypeError("Transformation type {} does not support".format(transformation.type))

    def _insert(self, target_point: Union[TargetPoint, TFMultiLayerPoint], insertion_objects: List[Callable]):
        if isinstance(target_point, TFMultiLayerPoint):
            self._shared_insert_layers(target_point.target_points, insertion_objects)
        elif isinstance(target_point, TFLayerWeight):
            weight_operations = [WeightOperations(target_point.weights_attr_name, insertion_objects)]
            self._insert_weight_operations(target_point.layer_name, weight_operations)
        elif isinstance(target_point, TFBeforeLayer):
            self._insert_layers_before(
                target_point.layer_name, target_point.instance_idx, target_point.input_port_id, insertion_objects
            )
        elif isinstance(target_point, TFAfterLayer):
            self._insert_layers_after(
                target_point.layer_name, target_point.instance_idx, target_point.output_port_id, insertion_objects
            )
        else:
            raise TypeError("Insertion transform does not support {} target point type".format(target_point.type))

    def _shared_insert_layers(self, target_points: List[TargetPoint], layers_to_insert: List[Callable]):
        functional_model = is_functional_model(self._model)
        if functional_model:
            for layer in self._model_config["input_layers"]:
                for tp in target_points:
                    if isinstance(tp, TFBeforeLayer) and tp.layer_name == layer[0]:
                        raise nncf.InternalError(f"Insertion before input layer: {tp.layer_name} is not supported")

        layer_configs = []
        for layer in layers_to_insert:
            config = tf.keras.utils.serialize_keras_object(layer)
            if functional_model:
                config["name"] = config["config"]["name"]
                config["inbound_nodes"] = []
                for i, tp in enumerate(target_points):
                    if isinstance(tp, TFAfterLayer):
                        config["inbound_nodes"].append([[tp.layer_name, tp.instance_idx, tp.output_port_id, {}]])
                    elif isinstance(tp, TFBeforeLayer):
                        idx, input_layer_cfg = self._find_layer_config(tp.layer_name)
                        inbound = input_layer_cfg["inbound_nodes"][tp.instance_idx][tp.input_port_id]
                        config["inbound_nodes"].append([[inbound[0], inbound[1], inbound[2], {}]])
                        self._model_config["layers"][idx]["inbound_nodes"][tp.instance_idx][tp.input_port_id] = [
                            config["name"],
                            i,
                            0,
                            inbound[3],
                        ]
                    else:
                        raise TypeError(
                            f"Insertion transform does not support {target_points[0].type} target point type"
                        )

            layer_configs.append(config)

        for config in layer_configs:
            for i, tp in enumerate(target_points):
                if functional_model and isinstance(tp, TFAfterLayer):
                    layer_out_ports = set()
                    replace_layer_name = config["name"]
                    for layer in self._model_config["layers"]:
                        inbound_nodes = layer["inbound_nodes"]
                        if layer["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
                            inbound_nodes = reformat_inbound_nodes_for_oplambda(inbound_nodes)

                        for inbound_node in inbound_nodes:
                            self._process_insertion_after(
                                inbound_node, tp.layer_name, tp.instance_idx, layer_out_ports, replace_layer_name, i
                            )

                    self._insert_after_model_outputs(
                        tp.layer_name, tp.instance_idx, layer_out_ports, replace_layer_name, i
                    )
                    if len(layer_out_ports) > 1:
                        raise nncf.InternalError(
                            "Insertion after layer ({}) with multiple ports is not supported".format(tp.layer_name)
                        )

            layer_name = target_points[0].layer_name
            self._insert_layer_after_sequential(layer_name, config)

    def _multi_insertion(self, target_point: TargetPoint, commands: List[TransformationCommand]):
        if not isinstance(target_point, TFLayer):
            raise TypeError(
                "Multiple insertion transform does not support {} target point type".format(target_point.type)
            )

        weight_operations = []
        for cmd in commands:
            if cmd.type != TransformationType.INSERT or cmd.target_point.type != TargetType.OPERATION_WITH_WEIGHTS:
                raise TypeError(
                    "Multiple insertion transform does not support command: "
                    "command type - {}; target point type - {}".format(cmd.type, cmd.target_point.type)
                )
            weight_operations.append(WeightOperations(cmd.target_point.weights_attr_name, cmd.insertion_objects))

        self._insert_weight_operations(target_point.layer_name, weight_operations)

    def _remove(self, target_point: TargetPoint):
        if isinstance(target_point, TFOperationWithWeights):
            target_layer_name = target_point.layer_name
            self._remove_weight_operation(
                target_layer_name, target_point.weights_attr_name, target_point.operation_name
            )
        else:
            raise TypeError("{} removal does not support".format(target_point.type))

    def _remove_weight_operation(self, layer_name: str, weights_attr_name: str, operation_name: str):
        _, layer_config = self._find_layer_config(layer_name)
        weights_operations = layer_config["config"]["weights_attr_operations"].pop(weights_attr_name)

        def find_weights_operation(operations, name):
            for op in operations:
                if op["config"]["name"] == name:
                    return op
            return None

        found = find_weights_operation(weights_operations, operation_name)
        weights_operations.remove(found)

        if weights_operations:
            layer_config["config"]["weights_attr_operations"][weights_attr_name] = weights_operations
        elif not layer_config["config"]["weights_attr_operations"]:
            self._replace_config(layer_name, layer_config["config"]["layer"])

    def _insert_weight_operations(self, layer_name: str, weight_operations: List[WeightOperations]):
        """
        Assigns the operations to be executed with a layer weight while accessing it.
        Any operation that have already been associated with the weight will be replaced with the
        new operation(s).
        :param layer_name: The name of the weighted layer.
        :param weight_operations: The operations to be executed with the weight
        """
        layer = self._get_layer(layer_name)
        wrapper = layer if isinstance(layer, NNCFWrapper) else NNCFWrapper(layer)

        for weights_attr_name, operations in weight_operations:
            for op in operations:
                wrapper.registry_weight_operation(weights_attr_name, op)

        self._replace(layer_name, wrapper)

    def _replace(self, layer_name: str, replace_layer):
        # NOTE: the change to the layer will only get propagated to self._model.layers
        # once the model is deserialized from config, i.e. not immediately.
        replace_layer_config = tf.keras.utils.serialize_keras_object(replace_layer)
        self._replace_config(layer_name, replace_layer_config)

    def _replace_config(self, layer_name: str, replace_layer_config: Dict):
        replace_layer_name = replace_layer_config["config"]["name"]
        if is_functional_model(self._model):
            if "name" not in replace_layer_config:
                replace_layer_config["name"] = replace_layer_name
            self._replace_functional(layer_name, replace_layer_config)
        else:
            self._replace_sequential(layer_name, replace_layer_config)

        self._update_layer_mapping(layer_name, replace_layer_name)

    def _replace_functional(self, layer_name: str, replace_layer_config: Dict):
        replace_layer_name = replace_layer_config["name"]
        for layer in self._model_config["layers"]:
            for inbound_node in layer["inbound_nodes"]:
                self._process_replacement(inbound_node, layer_name, replace_layer_name)

        self._replace_in_model_outputs(layer_name, replace_layer_name)

        idx, layer_config = self._find_layer_config(layer_name)
        replace_layer_config["inbound_nodes"] = layer_config["inbound_nodes"]
        self._model_config["layers"][idx] = replace_layer_config

    def _replace_sequential(self, layer_name: str, replace_layer_config: Dict):
        idx, _ = self._find_layer_config(layer_name)
        self._model_config["layers"][idx] = replace_layer_config

    def _insert_layers_before(self, layer_name: str, instance_idx: int, input_port_id: int, layers_to_insert: List):
        """
        Performs insertion before the (downstream) layer.

        :param layer_name: Name of the layer, before which to insert (downstream).
        :param instance_idx: Instance ID of the layer, before which to insert (downstream).
        :param input_port_id: Input port ID of the layer, before which to insert (downstream).
        :param layers_to_insert: List of the layers, which will be inserted into the graph.
        """
        functional_model = is_functional_model(self._model)

        if functional_model:
            for layer in self._model_config["input_layers"]:
                if layer_name == layer[0]:
                    raise nncf.InternalError("Insertion before input layer: {} is not supported".format(layer_name))

        layer_configs = []
        idx, downstream_layer_cfg = self._find_layer_config(layer_name)

        for layer in layers_to_insert:
            config = tf.keras.utils.serialize_keras_object(layer)
            if functional_model:
                config["name"] = config["config"]["name"]

                downstream_layer_inbound_nodes = downstream_layer_cfg["inbound_nodes"]
                if downstream_layer_cfg["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
                    downstream_layer_inbound_nodes = reformat_inbound_nodes_for_oplambda(downstream_layer_inbound_nodes)

                # Update config of the layer to insert
                inbound_node_info = copy.deepcopy(downstream_layer_inbound_nodes)[instance_idx][input_port_id]
                config["inbound_nodes"] = [[inbound_node_info[0], inbound_node_info[1], inbound_node_info[2], {}]]

                # Downstream layer config update
                if downstream_layer_cfg["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
                    downstream_layer_inbound_nodes[instance_idx][input_port_id][0] = config["name"]
                    downstream_layer_inbound_nodes[instance_idx][input_port_id][1] = 0
                    downstream_layer_inbound_nodes[instance_idx][input_port_id][2] = 0
                else:
                    self._model_config["layers"][idx]["inbound_nodes"][instance_idx][input_port_id] = [
                        config["name"],
                        0,
                        0,
                        {},
                    ]

            layer_configs.append(config)

        for config in layer_configs:
            self._model_config["layers"].insert(idx, config)

    def _insert_layers_after(self, layer_name: str, instance_idx: int, output_port_id: int, layers_to_insert: List):
        """
        Performs insertion after the (upstream) layer.

        :param layer_name: Name of the layer, after which to insert (upstream).
        :param instance_idx: Instance ID of the layer, after which to insert (upstream).
        :param input_port_id: Input port ID of the layer, after which to insert (upstream).
        :param layers_to_insert: List of the layers, which will be inserted into the graph.
        """
        functional_model = is_functional_model(self._model)

        layer_configs = []
        for layer in layers_to_insert:
            config = tf.keras.utils.serialize_keras_object(layer)
            if functional_model:
                config["name"] = config["config"]["name"]
                config["inbound_nodes"] = [[[layer_name, instance_idx, output_port_id, {}]]]
            layer_configs.append(config)

        for config in layer_configs:
            if functional_model:
                self._insert_layer_after_functional(layer_name, instance_idx, config)
            else:
                self._insert_layer_after_sequential(layer_name, config)

    def _insert_layer_after_functional(self, layer_name: str, instance_idx: int, layer_to_insert_config: Dict):
        """
        Performs insertion after the (upstream) layer (functional model).

        :param layer_name: Name of the layer, after which to insert (upstream).
        :param instance_idx: Instance ID of the Layer, after which to insert (upstream).
        :param layer_to_insert_config: Config of the layer, which is supposed to be inserted into the graph.
        """
        layer_out_ports = set()
        replace_layer_name = layer_to_insert_config["name"]  # new inbound_node name in the downstream layer

        for layer in self._model_config["layers"]:
            inbound_nodes = layer["inbound_nodes"]

            if layer["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
                inbound_nodes = reformat_inbound_nodes_for_oplambda(inbound_nodes)

            for inbound_node in inbound_nodes:
                self._process_insertion_after(
                    inbound_node, layer_name, instance_idx, layer_out_ports, replace_layer_name
                )

        self._insert_after_model_outputs(layer_name, instance_idx, layer_out_ports, replace_layer_name)
        if len(layer_out_ports) > 1:
            raise nncf.InternalError(
                "Insertion after layer ({}) with multiple ports is not supported".format(layer_name)
            )
        self._insert_layer_after_sequential(layer_name, layer_to_insert_config)

    def _insert_layer_after_sequential(self, layer_name: str, layer_configs):
        idx, _ = self._find_layer_config(layer_name)
        if idx is None:
            raise nncf.InternalError("Layer is not found: {}".format(layer_name))
        self._model_config["layers"].insert(idx + 1, layer_configs)

    @staticmethod
    def _process_insertion_after(
        connection_infos,
        layer_name: str,
        instance_idx: int,
        layer_out_ports: Set,
        replace_layer_name: str,
        insert_with_instance_idx: int = 0,
    ):
        if not isinstance(connection_infos[0], list):
            connection_infos = [connection_infos[0]]

        for connection_info in connection_infos:
            if connection_info[0] == layer_name:
                layer_out_ports.add(connection_info[2])
                if connection_info[1] == instance_idx:
                    connection_info[0] = replace_layer_name
                    connection_info[1] = insert_with_instance_idx

    def _insert_after_model_outputs(
        self,
        layer_name: str,
        instance_idx: int,
        layer_out_ports: Set,
        replace_layer_name: str,
        insert_with_instance_idx: int = 0,
    ):
        output_layers = self._model_config["output_layers"]
        if isinstance(output_layers, list):
            self._process_insertion_after(output_layers, layer_name, instance_idx, layer_out_ports, replace_layer_name)
        elif isinstance(output_layers, dict):
            for out_layers in output_layers.values():
                if isinstance(out_layers, list):
                    self._process_insertion_after(
                        [out_layers],
                        layer_name,
                        instance_idx,
                        layer_out_ports,
                        replace_layer_name,
                        insert_with_instance_idx,
                    )
                elif isinstance(out_layers, dict):
                    self._process_insertion_after(
                        list(out_layers.values()),
                        layer_name,
                        instance_idx,
                        layer_out_ports,
                        replace_layer_name,
                        insert_with_instance_idx,
                    )

    @staticmethod
    def _process_replacement(connection_infos, layer_name: str, replace_layer_name: str):
        if not isinstance(connection_infos[0], list):
            connection_infos = [connection_infos[0]]

        for connection_info in connection_infos:
            if connection_info[0] == layer_name:
                connection_info[0] = replace_layer_name

    def _replace_in_model_outputs(self, layer_name: str, replace_layer_name: str):
        output_layers = self._model_config["output_layers"]
        if isinstance(output_layers, list):
            self._process_replacement(output_layers, layer_name, replace_layer_name)
        elif isinstance(output_layers, dict):
            for out_layers in output_layers.values():
                if isinstance(out_layers, list):
                    self._process_replacement([out_layers], layer_name, replace_layer_name)
                elif isinstance(out_layers, dict):
                    self._process_replacement(list(out_layers.values()), layer_name, replace_layer_name)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import onnx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.tensor import Tensor
from nncf.onnx.graph.metatypes.groups import OPERATIONS_WITH_BIAS_REDUCED
from nncf.onnx.graph.node_utils import get_bias_value
from nncf.onnx.graph.node_utils import is_any_weight_quantized
from nncf.onnx.graph.node_utils import is_node_with_bias
from nncf.onnx.graph.transformations.command_creation import create_bias_correction_command
from nncf.onnx.graph.transformations.commands import ONNXInitializerUpdateCommand
from nncf.onnx.graph.transformations.commands import ONNXModelExtractionCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint
from nncf.onnx.statistics.collectors import get_mean_statistic_collector
from nncf.quantization.algorithms.fast_bias_correction.backend import FastBiasCorrectionAlgoBackend


class ONNXFastBiasCorrectionAlgoBackend(FastBiasCorrectionAlgoBackend):
    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> ONNXTargetPoint:
        return ONNXTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> ONNXInitializerUpdateCommand:
        return create_bias_correction_command(node, bias_value.data)

    @staticmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> ONNXModelExtractionCommand:
        return ONNXModelExtractionCommand(input_ids, output_ids)

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size, inplace)

    @staticmethod
    def get_sub_input_output_names(subgraph: onnx.ModelProto) -> Tuple[str, str]:
        return subgraph.graph.input[0].name, subgraph.graph.output[0].name

    @staticmethod
    def create_input_data(
        shape: Tuple[int], data: List[Tensor], input_name: str, channel_axis: int
    ) -> Dict[str, np.array]:
        blob = np.zeros(shape, dtype=data[0].data.dtype)
        for j, idx in enumerate(np.ndindex(blob.shape[channel_axis])):
            index = tuple(slice(None) if i != channel_axis else idx for i in range(blob.ndim))
            blob[index] = data[j].data
        input_data = {input_name: blob}
        return input_data

    @staticmethod
    def get_bias_value(node: NNCFNode, nncf_graph: NNCFGraph, model: onnx.ModelProto) -> Tensor:
        return Tensor(get_bias_value(node, model))

    @staticmethod
    def get_activation_port_ids_for_bias_node(node: NNCFNode) -> Tuple[int, int]:
        return 0, 0

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: str) -> Tensor:
        return Tensor(raw_data[output_name])

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_any_weight_quantized(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node) and node.metatype in OPERATIONS_WITH_BIAS_REDUCED

    @staticmethod
    def get_node_names_for_input_output_statistics(node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[str, str]:
        return node.node_name, node.node_name



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns.patterns import GraphPattern
from nncf.common.graph.patterns.patterns import IgnoredPatternNames
from nncf.common.utils.registry import Registry
from nncf.torch.graph.pattern_operations import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.torch.graph.pattern_operations import LINEAR_OPERATIONS

PT_IGNORED_PATTERNS = Registry("IGNORED_PATTERNS")


def _add_softmax_matmul(
    pattern: GraphPattern,
    matmul_aliases,
    reshape_squeeze_aliases,
    gather_aliases,
    transpose_aliases,
    concat_aliases,
) -> None:
    #       SOFTMAX  RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #           \              /
    #            \            /
    #             \          /
    #              \        /
    #               \      /
    #                MATMUL
    branch_matmul_nodes = reshape_squeeze_aliases + gather_aliases + transpose_aliases + concat_aliases
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: "softmax"})
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: matmul_aliases})
    matmul_branch_nodes = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "NON_PATTERN", GraphPattern.METATYPE_ATTR: branch_matmul_nodes}
    )
    pattern.add_edge(softmax, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


def _add_softmax_reshape_matmul(
    pattern: GraphPattern,
    matmul_aliases,
    reshape_squeeze_aliases,
    gather_aliases,
    transpose_aliases,
    concat_aliases,
) -> None:
    #       SOFTMAX
    #           \
    #            \
    #             \
    #             RESHAPE   RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #                 \                 /
    #                  \               /
    #                   \             /
    #                    \           /
    #                     \         /
    #                      \       /
    #                        MATMUL
    branch_matmul_nodes = reshape_squeeze_aliases + gather_aliases + transpose_aliases + concat_aliases
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: "softmax"})
    reshape = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "RESHAPE", GraphPattern.METATYPE_ATTR: reshape_squeeze_aliases}
    )
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: matmul_aliases})
    matmul_branch_nodes = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER", GraphPattern.METATYPE_ATTR: branch_matmul_nodes}
    )
    pattern.add_edge(softmax, reshape)
    pattern.add_edge(reshape, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)
    return pattern


@PT_IGNORED_PATTERNS.register(IgnoredPatternNames.MULTIHEAD_ATTENTION_OUTPUT)
def create_multihead_attention_output() -> GraphPattern:
    matmul_aliases = ["linear", "addmm", "matmul", "bmm", "mm", "baddbmm", "__matmul__"]
    reshape_squeeze_aliases = [
        "reshape",
        "view",
        "flatten",
        "unsqueeze",
        "squeeze",
        "unbind",
    ]
    gather_aliases = ["gather", "index_select", "where", "index_select", "__getitem__"]
    transpose_aliases = ["transpose", "permute", "transpose_"]
    concat_aliases = ["cat", "stack"]

    pattern = GraphPattern()
    _add_softmax_matmul(
        pattern,
        matmul_aliases=matmul_aliases,
        reshape_squeeze_aliases=reshape_squeeze_aliases,
        gather_aliases=gather_aliases,
        transpose_aliases=transpose_aliases,
        concat_aliases=concat_aliases,
    )
    _add_softmax_reshape_matmul(
        pattern,
        matmul_aliases=matmul_aliases,
        reshape_squeeze_aliases=reshape_squeeze_aliases,
        gather_aliases=gather_aliases,
        transpose_aliases=transpose_aliases,
        concat_aliases=concat_aliases,
    )
    return pattern


# pylint:disable=too-many-statements
@PT_IGNORED_PATTERNS.register(IgnoredPatternNames.SE_BLOCK)
def create_se_block() -> GraphPattern:
    MEAN_OPERATIONS = {
        GraphPattern.LABEL_ATTR: "REDUCE_MEAN",
        GraphPattern.METATYPE_ATTR: ["avg_pool2d", "adaptive_avg_pool2d", "avg_pool3d", "adaptive_avg_pool3d", "mean"],
        GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
    }
    SYGMOID_OPERATIONS = {
        GraphPattern.LABEL_ATTR: "SIGMOID",
        GraphPattern.METATYPE_ATTR: ["sigmoid", "hardsigmoid"],
    }
    MUL_OPERATION = {
        GraphPattern.LABEL_ATTR: "MUL",
        GraphPattern.METATYPE_ATTR: "__mul__",
        GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
    }

    def get_se_block_pattern() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, linear_node_1)
        pattern.add_edge(linear_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    def get_se_block_with_bias_pattern() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_1 = pattern.add_node(label="ADD_BIAS", type=["__add__", "__sub__"])
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_2 = pattern.add_node(label="ADD_BIAS", type=["__add__", "__sub__"])
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, linear_node_1)
        pattern.add_edge(linear_node_1, add_node_1)
        pattern.add_edge(add_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, add_node_2)
        pattern.add_edge(add_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    RESHAPE_NODES = {
        GraphPattern.LABEL_ATTR: "RESHAPE",
        GraphPattern.METATYPE_ATTR: ["reshape", "view", "flatten", "unsqueeze"],
    }

    def get_se_block_with_reshape() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        reshape_node_1 = pattern.add_node(**RESHAPE_NODES)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        reshape_node_2 = pattern.add_node(**RESHAPE_NODES)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, reshape_node_1)
        pattern.add_edge(reshape_node_1, linear_node_1)
        pattern.add_edge(linear_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, reshape_node_2)
        pattern.add_edge(reshape_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    def get_se_block_with_bias_and_reshape() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        reshape_node_1 = pattern.add_node(**RESHAPE_NODES)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_1 = pattern.add_node(label="ADD_BIAS", type=["__add__", "__sub__"])
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_2 = pattern.add_node(label="ADD_BIAS", type=["__add__", "__sub__"])
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        reshape_node_2 = pattern.add_node(**RESHAPE_NODES)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, reshape_node_1)
        pattern.add_edge(reshape_node_1, linear_node_1)
        pattern.add_edge(linear_node_1, add_node_1)
        pattern.add_edge(add_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, add_node_2)
        pattern.add_edge(add_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, reshape_node_2)
        pattern.add_edge(reshape_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    main_pattern = GraphPattern()
    main_pattern.add_pattern_alternative(get_se_block_pattern())
    main_pattern.add_pattern_alternative(get_se_block_with_bias_pattern())
    main_pattern.add_pattern_alternative(get_se_block_with_reshape())
    main_pattern.add_pattern_alternative(get_se_block_with_bias_and_reshape())
    return main_pattern



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections.abc import Iterator
from typing import Dict, List, Tuple

from nncf.common.graph.graph import NNCFNode
from nncf.experimental.tensor.tensor import Tensor


class LayerwiseIterator(Iterator, ABC):
    """
    An abstract base class for iterating through the layers of a model.
    """

    @abstractmethod
    def __next__(self) -> Tuple[NNCFNode, Dict[int, List[Tensor]]]:
        """
        Returns the next node and its associated tensors in the iteration.

        :return: The next node and its associated tensors in the iteration.
        :raises StopIteration: When there are no more elements to iterate over.
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from typing import Dict, List, Tuple

import tensorflow as tf

from nncf import NNCFConfig
from nncf.common.compression import BaseCompressionAlgorithmController
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.node_selector import PruningNodeSelector
from nncf.common.pruning.statistics import PrunedLayerSummary
from nncf.common.pruning.structs import PrunedLayerInfoBase
from nncf.common.pruning.utils import get_output_channels
from nncf.common.pruning.utils import is_prunable_depthwise_conv
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.schemata.defaults import PRUNE_BATCH_NORMS
from nncf.config.schemata.defaults import PRUNE_DOWNSAMPLE_CONVS
from nncf.config.schemata.defaults import PRUNE_FIRST_CONV
from nncf.config.schemata.defaults import PRUNING_INIT
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.converter import TFModelConverterFactory
from nncf.tensorflow.graph.metatypes.keras_layers import TFBatchNormalizationLayerMetatype
from nncf.tensorflow.graph.model_transformer import TFModelTransformer
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import get_layer_identifier
from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.pruning.operations import TF_PRUNING_OPERATOR_METATYPES
from nncf.tensorflow.pruning.operations import TFElementwisePruningOp
from nncf.tensorflow.pruning.operations import TFIdentityMaskForwardPruningOp
from nncf.tensorflow.pruning.tensor_processor import TFNNCFPruningTensorProcessor
from nncf.tensorflow.pruning.utils import get_filter_axis
from nncf.tensorflow.pruning.utils import get_filters_num
from nncf.tensorflow.sparsity.magnitude.operation import BinaryMask
from nncf.tensorflow.sparsity.utils import strip_model_from_masks
from nncf.tensorflow.tensor import TFNNCFTensor


class PrunedLayerInfo(PrunedLayerInfoBase):
    def __init__(self, node_name: NNCFNodeName, layer_name: str, node_id: int, is_depthwise: bool):
        super().__init__(node_name, node_id, is_depthwise)
        self.layer_name = layer_name


class BasePruningAlgoBuilder(TFCompressionAlgorithmBuilder):
    """
    Determines which modifications should be made to the original model in
    order to enable pruning during fine-tuning.
    """

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        params = self._algo_config.get("params", {})
        self._params = params

        self._ignore_frozen_layers = True
        self._prune_first = params.get("prune_first_conv", PRUNE_FIRST_CONV)
        self._prune_batch_norms = params.get("prune_batch_norms", PRUNE_BATCH_NORMS)
        self._prune_downsample_convs = params.get("prune_downsample_convs", PRUNE_DOWNSAMPLE_CONVS)

        self._prunable_types = self._get_op_types_of_pruned_layers()
        types_of_grouping_ops = self._get_types_of_grouping_ops()
        self._pruning_node_selector = PruningNodeSelector(
            TF_PRUNING_OPERATOR_METATYPES,
            self._prunable_types,
            types_of_grouping_ops,
            self.ignored_scopes,
            self.target_scopes,
            self._prune_first,
            self._prune_downsample_convs,
        )

        self._pruned_layer_groups_info = None
        self._graph = None
        self._op_names = []

    def apply_to(self, model: tf.keras.Model) -> tf.keras.Model:
        """
        Adds pruning masks to the model.

        :param model: The original uncompressed model.
        :return: The model with pruning masks.
        """
        transformer = TFModelTransformer(model)
        transformation_layout = self.get_transformation_layout(model)
        return transformer.transform(transformation_layout)

    def get_transformation_layout(self, model: tf.keras.Model) -> TFTransformationLayout:
        """
        Computes necessary model transformations (pruning mask insertions) to enable pruning.

        :param model: The original uncompressed model.
        :return: The instance of the `TransformationLayout` class containing
            a list of pruning mask insertions.
        """
        converter = TFModelConverterFactory.create(model)
        self._graph = converter.convert()

        check_scopes_in_graph(self._graph, self.ignored_scopes, self.target_scopes, self.validate_scopes)

        groups_of_nodes_to_prune = self._pruning_node_selector.create_pruning_groups(self._graph)

        transformations = TFTransformationLayout()
        shared_layers = set()

        self._pruned_layer_groups_info = Clusterization[PrunedLayerInfo](lambda x: x.layer_name)

        for i, group in enumerate(groups_of_nodes_to_prune.get_all_clusters()):
            group_minfos = []
            for node in group.elements:
                layer_name = get_layer_identifier(node)
                layer = model.get_layer(layer_name)
                group_minfos.append(
                    PrunedLayerInfo(node.node_name, layer_name, node.node_id, is_prunable_depthwise_conv(node))
                )

                # Add output_mask to elements to run mask_propagation
                # and detect spec_nodes that will be pruned.
                # It should be done for all elements of shared layer.
                node.attributes["output_mask"] = TFNNCFTensor(tf.ones(get_output_channels(node)))
                if layer_name in shared_layers:
                    continue
                if node.is_shared():
                    shared_layers.add(layer_name)
                # Check that we need to prune weights in this op
                assert self._is_pruned_layer(layer)
                nncf_logger.debug(f"Will prune the weights for the layer: {layer_name}")

                _, layer_info = converter.get_layer_info_for_node(node.node_name)
                for weight_def in node.metatype.weight_definitions:
                    transformations.register(
                        self._get_insertion_command_binary_mask(layer_info.layer_name, weight_def.weight_attr_name)
                    )
                if (
                    node.metatype.bias_attr_name is not None
                    and getattr(layer, node.metatype.bias_attr_name) is not None
                ):
                    transformations.register(
                        self._get_insertion_command_binary_mask(layer_info.layer_name, node.metatype.bias_attr_name)
                    )

            cluster = Cluster[PrunedLayerInfo](i, group_minfos, [n.node_id for n in group.elements])
            self._pruned_layer_groups_info.add_cluster(cluster)

        # Propagating masks across the graph to detect spec_nodes that will be pruned
        mask_propagator = MaskPropagationAlgorithm(
            self._graph, TF_PRUNING_OPERATOR_METATYPES, TFNNCFPruningTensorProcessor
        )
        mask_propagator.mask_propagation()

        # Add masks for all spec modules, because prunable batchnorm layers can be determined
        # at the moment of mask propagation
        types_spec_layers = [TFBatchNormalizationLayerMetatype] if self._prune_batch_norms else []

        spec_nodes = self._graph.get_nodes_by_metatypes(types_spec_layers)
        for spec_node in spec_nodes:
            layer_name = get_layer_identifier(spec_node)
            layer = model.get_layer(layer_name)
            if spec_node.attributes["output_mask"] is None:
                # Skip elements that will not be pruned
                continue
            if layer_name in shared_layers:
                continue
            if spec_node.is_shared():
                shared_layers.add(layer_name)
            nncf_logger.debug(f"Will prune the weights for the layer: {layer_name}")

            _, layer_info = converter.get_layer_info_for_node(spec_node.node_name)
            for weight_def in spec_node.metatype.weight_definitions:
                if (
                    spec_node.metatype is TFBatchNormalizationLayerMetatype
                    and not layer.scale
                    and weight_def.weight_attr_name == "gamma"
                ):
                    nncf_logger.debug(
                        "Fused gamma parameter encountered in BatchNormalization layer. "
                        "Won't add a pruning mask to it."
                    )
                    continue

                transformations.register(
                    self._get_insertion_command_binary_mask(layer_info.layer_name, weight_def.weight_attr_name)
                )
            transformations.register(
                self._get_insertion_command_binary_mask(layer_info.layer_name, spec_node.metatype.bias_attr_name)
            )
        return transformations

    def initialize(self, model: tf.keras.Model) -> None:
        pass

    def _get_insertion_command_binary_mask(self, layer_name: str, attr_name: str) -> TFInsertionCommand:
        op_name = self._get_pruning_operation_name(layer_name, attr_name)
        self._op_names.append(op_name)

        return TFInsertionCommand(
            target_point=TFLayerWeight(layer_name, attr_name),
            callable_object=BinaryMask(op_name),
            priority=TransformationPriority.PRUNING_PRIORITY,
        )

    @staticmethod
    def _get_bn_for_node(node: NNCFNode, bn_nodes: List[NNCFNode]) -> Tuple[bool, List[NNCFNode]]:
        is_finished = False
        propagating_ops = [
            op_name
            for meta_op in [TFIdentityMaskForwardPruningOp, TFElementwisePruningOp]
            for op_name in meta_op.get_all_op_aliases()
        ]
        if node.node_type == "BatchNormalization":
            is_finished = True
            bn_nodes.append(node)
        elif node.node_type not in propagating_ops:
            is_finished = True
        return is_finished, bn_nodes

    def _get_related_batchnorms(self, layer_name: str, group: Cluster, graph: NNCFGraph) -> List[NNCFNode]:
        """
        Returns List of batchnorm elements related to the layer.
        Note: Single node per layer for shared bactchnorm layers
        """
        layer_nodes = [node_ for node_ in group.elements if node_.layer_name == layer_name]
        bn_nodes = []
        bn_layer_names = []
        for layer_node in layer_nodes:
            for next_node in graph.get_next_nodes(layer_node):
                for bn_node in graph.traverse_graph(next_node, self._get_bn_for_node):
                    bn_layer_name = get_layer_identifier(bn_node)
                    if bn_layer_name not in bn_layer_names:
                        bn_layer_names.append(bn_layer_name)
                        bn_nodes.append(bn_node)
        return bn_nodes

    def _is_pruned_layer(self, layer: tf.keras.layers.Layer) -> bool:
        """
        Return whether this layer should be pruned or not.
        """
        raise NotImplementedError

    def _get_op_types_of_pruned_layers(self) -> List[str]:
        """
        Returns list of operation types that should be pruned.
        """
        raise NotImplementedError

    def _get_types_of_grouping_ops(self) -> List[str]:
        raise NotImplementedError

    def _get_pruning_operation_name(self, layer_name: str, weight_attr_name: str) -> str:
        return f"{layer_name}_{weight_attr_name}_pruning_binary_mask"


@api()
class BasePruningAlgoController(BaseCompressionAlgorithmController, ABC):
    """
    Base class for TF pruning algorithm controllers.
    """

    def __init__(
        self,
        target_model: tf.keras.Model,
        op_names: List[str],
        prunable_types: List[str],
        pruned_layer_groups_info: Clusterization[PrunedLayerInfo],
        config,
    ):
        super().__init__(target_model)
        self._op_names = op_names
        self._prunable_types = prunable_types
        self.config = config
        self.pruning_config = extract_algo_specific_config(config, "filter_pruning")
        params = self.pruning_config.get("params", {})
        self.pruning_init = self.pruning_config.get("pruning_init", PRUNING_INIT)
        self.pruning_level = self.pruning_init
        self._pruned_layer_groups_info = pruned_layer_groups_info
        self.prune_flops = False
        self._check_pruning_level(params)
        self._num_of_sparse_elements_by_node = None

    def freeze(self):
        raise NotImplementedError

    def set_pruning_level(self, pruning_level: float):
        raise NotImplementedError

    def step(self, next_step):
        pass

    def _check_pruning_level(self, params):
        """
        Check that set only one of pruning target params
        """
        pruning_target = params.get("pruning_target", None)
        pruning_flops_target = params.get("pruning_flops_target", None)
        if pruning_target and pruning_flops_target:
            raise ValueError("Only one parameter from 'pruning_target' and 'pruning_flops_target' can be set.")
        if pruning_flops_target:
            self.prune_flops = True

    def _calculate_num_of_sparse_elements_by_node(self) -> Dict[NNCFNodeName, int]:
        """Returns the number of sparse elements per node. Take into account names ('^') for the shared ops."""
        if self._num_of_sparse_elements_by_node is None:
            self._calculate_pruned_layers_summary()

        retval = {}
        for group in self._pruned_layer_groups_info.get_all_clusters():
            for node in group.elements:
                retval[node.node_name] = self._num_of_sparse_elements_by_node[node.layer_name]
        return retval

    def _calculate_pruned_layers_summary(self) -> List[PrunedLayerSummary]:
        pruning_levels = []
        mask_names = []
        weights_shapes = []
        mask_shapes = []
        self._num_of_sparse_elements_by_node = {}
        for wrapped_layer, weight_attr, op_name in get_nncf_operations(self._model, self._op_names):
            mask = wrapped_layer.ops_weights[op_name.name]["mask"]
            mask_names.append(mask.name)
            weights_shapes.append(list(mask.shape))
            reduce_axes = list(range(len(mask.shape)))
            filter_axis = get_filter_axis(wrapped_layer, weight_attr)
            if filter_axis == -1:
                filter_axis = reduce_axes[filter_axis]
            reduce_axes.remove(filter_axis)
            filter_mask = tf.reduce_max(tf.cast(mask, tf.int32), axis=reduce_axes, keepdims=True)
            mask_shapes.append(list(filter_mask.shape))
            filters_number = get_filters_num(wrapped_layer)
            pruned_filters_number = filters_number - tf.reduce_sum(filter_mask)
            pruning_levels.append(pruned_filters_number / filters_number)
            pruned_filter_number = filters_number - tf.reduce_sum(filter_mask)
            self._num_of_sparse_elements_by_node[wrapped_layer.name] = pruned_filter_number.numpy()

        pruning_levels = tf.keras.backend.batch_get_value(pruning_levels)
        mask_pruning = list(zip(mask_names, weights_shapes, mask_shapes, pruning_levels))

        pruned_layers_summary = []
        for mask_name, weights_shape, mask_shape, pruning_level in mask_pruning:
            pruned_layers_summary.append(PrunedLayerSummary(mask_name, weights_shape, mask_shape, pruning_level))

        return pruned_layers_summary

    def strip_model(self, model: tf.keras.Model, do_copy: bool = False) -> tf.keras.Model:
        # Transform model for pruning creates copy of the model.
        return strip_model_from_masks(model, self._op_names)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.statistics import NNCFStatistics
from nncf.tensorflow.callbacks.statistics_callback import StatisticsCallback


class PruningStatisticsCallback(StatisticsCallback):
    """
    Callback for logging pruning compression statistics to tensorboard and stdout.
    """

    def _prepare_for_tensorboard(self, stats: NNCFStatistics):
        base_prefix = "2.compression/statistics"
        detailed_prefix = "3.compression_details/statistics"

        ms = stats.filter_pruning.model_statistics
        tensorboard_stats = {
            f"{base_prefix}/algo_current_pruning_level": stats.filter_pruning.current_pruning_level,
            f"{base_prefix}/model_FLOPS_pruning_level": ms.flops_pruning_level,
            f"{base_prefix}/model_params_pruning_level": ms.params_pruning_level,
            f"{base_prefix}/model_filters_pruning_level": ms.filter_pruning_level,
        }

        for ls in ms.pruned_layers_summary:
            layer_name, pruning_level = ls.name, ls.filter_pruning_level
            tensorboard_stats[f"{detailed_prefix}/{layer_name}/pruning_level"] = pruning_level

        return tensorboard_stats



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import torch
from torch import nn

from nncf.torch.dynamic_graph.context import TracingContext


class KnowledgeDistillationLossHandler(nn.Module):
    """
    Encapsulates knowledge distillation logic. Controls knowledge distillation loss calculation. Notice that knowledge
    distillation loss is computed between results of original model and compressed model inferences only with latest
    inputs. And storages loss values in context at storage_device for further access. Such complex method of storage
    is required for DataParallel model replication logic.
    """

    KD_LOSS_STORAGE_NAME = "kd_loss"
    KD_STORAGE_DEVICE = "kd_storage_device"

    def __init__(
        self, context: TracingContext, kd_original_model: nn.Module, calculate_kd_loss_fn, storage_device: torch.device
    ):
        super().__init__()
        self._compressed_context = context
        self._kd_original_model = kd_original_model
        self._calculate_kd_loss_fn = calculate_kd_loss_fn
        self._compressed_context.register_global_buffer(self.KD_LOSS_STORAGE_NAME, [])
        self._compressed_context.register_global_buffer(self.KD_STORAGE_DEVICE, storage_device)

    def zero_kd_loss(self):
        """
        Frees storage space for further next iteration loss value storage.
        """
        self._compressed_context.global_buffer_store[self.KD_LOSS_STORAGE_NAME] = []

    def get_kd_loss(self) -> List[torch.Tensor]:
        if len(self._compressed_context.global_buffer_store[self.KD_LOSS_STORAGE_NAME]) == 0:
            return [torch.zeros([], device=self._compressed_context.global_buffer_store[self.KD_STORAGE_DEVICE])]
        return self._compressed_context.global_buffer_store[self.KD_LOSS_STORAGE_NAME]

    def forward(self, compressed_model_outputs, *args, **kwargs):
        """
        Infers kd original model with latest NNCFNetwork forward inputs (*args, **kwargs) and computes distillation loss
        between results of kd original model forward and compressed model forward (inputs). Then stores loss values
        in context at storage device.

        :param compressed_model_outputs: Results of compressed model forward used for
        knowledge distillation loss calculations.
        """
        self.zero_kd_loss()
        with torch.no_grad():
            original_model_outputs = self._kd_original_model(*args, **kwargs)
        kd_loss = self._calculate_kd_loss_fn(compressed_model_outputs, original_model_outputs)
        if kd_loss is not None:
            self._compressed_context.global_buffer_store[self.KD_LOSS_STORAGE_NAME].append(
                kd_loss.to(self._compressed_context.global_buffer_store[self.KD_STORAGE_DEVICE])
            )
